
##########################################################################
####################   Quatroid Properties    ############################
##########################################################################

#Extracts all triples in Q
function Triples(Q)
	filter(x->length(x)==3,Q)
end

#Extracts all sextuples in Q
function Sextuples(Q)
	filter(x->length(x)==6,Q)
end

#Checks if Bezouts first weak criterion is satisfied
function BezoutWeak1(Q)
	T = Triples(Q)
	S = Sextuples(Q)
	for I in T
		for J in S
			if length(I ∩ J)>2
				return false
			end
		end
	end

	return(true)
end

#Checks if Bezouts second weak criterion is satisfied
function BezoutWeak2(Q)
	T = Triples(Q)
	for I1 in T
		for I2 in T
			if length(I1 ∩ I2)==2 #if the lines intersect in 2 points
				Line = filter(x->x ∈ I1 ∪ I2,1:8)
				ImpliedIndices = [[1,2,3],[1,2,4],[1,3,4],[2,3,4]]
				TriplesImplied = [Line[I] for I in ImpliedIndices]
				for TI in TriplesImplied
					if TI ∉ T
						return(false)
					end
				end
			end
		end
	end
	return(true)
end

#Checks if Bezouts third weak criterion is satisfied
function BezoutWeak3(Q)
	S = Sextuples(Q)
	for J1 in S
		for J2 in S
			if length(J1 ∩ J2)==5 #if the conics intersect in 5 pts
				Conic = filter(x->x ∈ J1 ∪ J2,1:8)
				IndicesImplied =[[1,2,3,4,5,6],[1,2,3,4,5,7],[1,2,3,4,6,7],[1,2,3,5,6,7],[1,2,4,5,6,7],[1,3,4,5,6,7],[2,3,4,5,6,7]]
				SextuplesImplied = [Conic[J] for J in IndicesImplied]
				for C in SextuplesImplied
					if C ∉ S
						return(false)
					end
				end
			end
		end
	end
	return(true)
end

#Checks if Bezout's weak criteria are satisfied
function BezoutWeak(Q)
	if BezoutWeak1(Q) && BezoutWeak2(Q) && BezoutWeak3(Q)
		return(true)
	else
		return(false)
	end
end

#Closes the triples with respect to Bezout Weak 2
function BezoutWeakClosure1(Q)
	Qnew=copy(Q)
	T = Triples(Qnew)
	for I1 in T
		for I2 in T
			if length(I1 ∩ I2)==2
				Line = filter(x->x ∈ I1 ∪ I2,1:8)
				ImpliedIndices = [[1,2,3],[1,2,4],[1,3,4],[2,3,4]]
				TriplesImplied = [Line[I] for I in ImpliedIndices]
				for TI in TriplesImplied
					push!(Qnew,TI)
				end
			end
		end
	end
	Qnew = sort(unique(Qnew))
	if Q == Qnew
		return(Q)
	else
		return(BezoutWeakClosure1(Qnew))
	end
end

#Closes the sextuples with respect to Bezout Weak 3
function BezoutWeakClosure2(Q)
	Qnew=copy(Q)
	S = Sextuples(Qnew)
	for J1 in S
		for J2 in S
			if length(J1 ∩ J2)==5
				Conic = filter(x->x ∈ J1 ∪ J2,1:8)
				ImpliedIndices =[[1,2,3,4,5,6],[1,2,3,4,5,7],[1,2,3,4,6,7],[1,2,3,5,6,7],[1,2,4,5,6,7],[1,3,4,5,6,7],[2,3,4,5,6,7]]
				SextuplesImplied = [Conic[J] for J in ImpliedIndices]
				for SI in SextuplesImplied
					push!(Qnew,SI)
				end
			end
		end
	end
	Qnew = sort(unique(Qnew))
	if Q == Qnew
		return(Q)
	else
		return(BezoutWeakClosure2(Qnew))
	end
end

#Closes the triples and sextuples with respect to Bezout Weak 2 and 3
function BezoutClosure(Q)
	Q=BezoutWeakClosure1(Q)
	Q=BezoutWeakClosure2(Q)
	return(Q)
end

#Checks if Bezouts first strong criterion is satisfied
function BezoutStrong1(Q)
	T = Triples(Q)
	for I1 in T
		for I2 in T
			if 1<length(I1 ∩ I2)<3
				return(false)
			end
		end
	end
	return(true)
end

#Checks if Bezouts second strong criterion is satisfied
function BezoutStrong2(Q)
	S = Sextuples(Q)
	for J1 in S
		for J2 in S
			if length(J1∩J2)==5
				return(false)
			end
		end
	end
	return(true)
end


function BezoutStrong(Q)
	if BezoutStrong1(Q) && BezoutStrong2(Q)
		return(true)
	else
		return(false)
	end
end

function isBezoutian(Q)
	if BezoutWeak(Q) && BezoutStrong(Q)
		return(true)
	else
		return(false)
	end
end

#Given a quatroid, returns the index of the matroidal part
function MatroidNumber(Q)
	T = Triples(Q)
	AllMatroids=GenerateAllMatroids()
	FF = findfirst(x->x==T,AllMatroids)
	if FF == nothing
		T = sort([sort(t) for t in T])
		MI = minimalImage(Triples(Q))
		PossibleMatroids=filter(x->length(x)==length(MI),AllMatroids)
		ff = findfirst(x->x==MI,[minimalImage(M) for M in PossibleMatroids])
		FF = findfirst(x->x==PossibleMatroids[ff],AllMatroids)
		if FF == nothing
			println("Error: ",Q)
		end
	end
    return(FF)
end

# Returns a signature of the quatroid:
#  the matroid index
#  the number of triples
#  the number of sextuples
#  a sorted 8-tuple of pairs, where the i-th part is (a,b) where
#     a is the number of triples containing i
#     b is the number of sextuples containing i
function QuatroidSignature(Q)
	T = Triples(Q)
	S = Sextuples(Q)
	MN = MatroidNumber(Q)
	LineConicInvolvement = sort([[length(findall(x->i ∈ x,T)),length(findall(x->i∈x,S))] for i in 1:8])
	return([MN,length(T),length(S),LineConicInvolvement])
end

function QuatroidNumber(Q;sd=false)
	if sd==true
		Q=ShortToQuatroid(Q)
	end

	AllQuatroids = GenerateAllCandidateQuatroids()

	QS = QuatroidSignature(Q)
	QSall = QuatroidSignatures()
	QN = findfirst(x->QSall[x]==QS,1:length(QSall))

	if QN == nothing
		println("Cannot find quatroid signature in list *******")
		println(Q)
		println("*************")
	end
	return(QN)
end

## Produce a file containing short descriptions of Q=(I,J)
# e.g. if 123,124,134,234 are in I, replace them with 1234 in I'
# if 123456,123457,...,234567 are in J, replace them with 1234567 in J'
# return the pair (I',J')
function shortDescription(Q;Conics=[],Lines=[])
    Triples = filter(x->length(x)==3,Q)
    Sextuples = filter(x->length(x)==6,Q)
    newQ = []
    if length(Conics)==0
        if length(Sextuples)>0
            push!(Conics,Sextuples[1])
        end
    else
        if length(Sextuples)>0
            push!(Conics,Sextuples[1])
        end
    end
    if length(Lines)==0
        if length(Triples)>0
            push!(Lines,Triples[1])
        end
    else
        if length(Triples)>0
            push!(Lines,Triples[1])
        end
    end
    for T in Triples
        withinaline=false
        for i in 1:length(Lines)
            l=Lines[i]
        #    println(length(T ∩ l))
            if length(T ∩ l)>1
                Lines[i]=sort(unique(l ∪ T))
                withinaline=true
            end
        end
        if withinaline==false
            push!(newQ,T)
        end
    end
    for S in Sextuples
        withinaconic=false
        for j in 1:length(Conics)
            Q=Conics[j]
            if length(S ∩ Q)>4
                Conics[j]=sort(unique(Q ∪ S))
                withinaconic=true
            end
        end
        if withinaconic==false
            push!(newQ,S)
        end
    end
    if length(newQ)>0
        return(shortDescription(newQ;Conics=Conics,Lines=Lines))
    else
        return([Lines,Conics])
    end
end

# Changes the short description back to the quatroid format
function ShortToQuatroid(Q)
	Qnew = []
	for l in Q[1]
		ltriples = filter(x->issubset(x,l), collect(combinations(1:8,3)))
		for t in ltriples
			push!(Qnew,t)
		end
	end
	for q in Q[2]
		qsextuples = filter(x->issubset(x,q),collect(combinations(1:8,6)))
		for s in qsextuples
			push!(Qnew,s)
		end
	end
	Q=[sort(q) for q in Qnew]
	sort(Q)
end

#An exhaustive quatroid is one which has each point in at least one figure
function isExhaustive(Q)
	involved = 0
	for i in 1:8
		inv = false
		for q in Q
			if inv==false
				if i ∈ q
					inv=true
				end
			end
		end
		if inv==true
			involved=involved+1
		end
	end
	if involved==8
		return(true)
	else
		return(false)
	end
end

#Checks if Q has a triple meeting a sextuple at one point
function ContainedIn10(Q)
	Triples = filter(x->length(x)==3,Q)
	Sextuples = filter(x->length(x)==6,Q)
	for q in Sextuples
		for t in Triples
			if length(q ∩ t) ==1
				return(true)
			end
		end
	end
	return(false)
end

#Checks if Q has two triples meeting at a point and the triple of that union's complement
function ContainedIn77(Q)
	Triples = filter(x->length(x)==3,Q)
	for I1 in Triples
		for I2 in Triples
			if length(I1 ∩ I2)==1
				for I3 in Triples
					if length(I1 ∩ I3)==0 && length(I2 ∩ I3)==0
						return(true)
					end
				end
			end
		end
	end
	return(false)
end

#If a Bezoutian quatroid is exhaustive and does not contain Q77 or Q10, it has
# a reduced base locus. This is a sufficient condition, not necessary.
function ReducedBaseLocus(Q)
	if isBezoutian(Q)==false
		return(nothing)
	end
	if isExhaustive(Q)==false
		return(nothing)
	end
	if ContainedIn77(Q)==false && ContainedIn10(Q)==false
		return(true)
	else
		return(nothing)
	end
end

function mQ(Q;verbose=false)
	if isBezoutian(Q)==false
		return(Inf)
	end
	T = Triples(Q)
	S = Sextuples(Q)
	m = 2*length(T) + 2*length(S)
	for I in T
		for J in S
			if length(I∩J)==1
				m=m-2
			end
		end
	end
	for I1 in T
		for I2 in T
			if I1<I2 #ensures pairs are counted once
				if length(I1 ∩ I2)==0
					m=m-1
				end
			end
		end
	end
	for I1 in T
		for I2 in T
			for I3 in T
				if I1<I2 && I2 <I3
					if length(unique(vcat([I1,I2,I3]...)))==8
						m=m-1
					end
				end
			end
		end
	end
	return(m)
end
